
(define-type FAE
  [num (n number?)]
  [add (lhs FAE?)
       (rhs FAE?)]
  [sub (lhs FAE?)
       (rhs FAE?)]
  [id (name symbol?)]
  [fun (param symbol?)
       (body FAE?)]
  [app (fun-expr FAE?)
       (arg-expr FAE?)]
  [if0 (test FAE?)
       (then FAE?)
       (else FAE?)])

(define-type CFAE
  [cnum (n number?)]
  [cadd (lhs CFAE?)
        (rhs CFAE?)]
  [csub (lhs CFAE?)
        (rhs CFAE?)]
  [cid (pos number?)]
  [cfun (body CFAE?)]
  [capp (fun-expr CFAE?)
       (arg-expr CFAE?)]
  [cif0 (test CFAE?)
        (then CFAE?)
        (else CFAE?)])

(define-type CFAE-Value
  [numV (n number?)]
  [closureV (body CFAE?)
            (ds DefrdSub?)])

(define-type CDefrdSub
  [mtCSub]
  [aCSub (name symbol?)
         (rest CDefrdSub?)])

(define-type CFAE-Cont
  [mtK]
  [addSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
  [doAddK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
  [subSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
  [doSubK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
  [appArgK (arg-expr CFAE?)
           (ds DefrdSub?)
           (k CFAE-Cont?)]
  [doAppK (fun-val CFAE-Value?)
          (k CFAE-Cont?)]
  [doIfK (then-expr CFAE?)
         (else-expr CFAE?)
         (ds DefrdSub?)
         (k CFAE-Cont?)])

;; A DefrdSub is a list of CFAE-Value
(define (DefrdSub? x) true)

;; ----------------------------------------

;; parse : S-expr -> FAE
(define (parse sexp)
  (cond
    [(number? sexp) (num sexp)]
    [(symbol? sexp) (id sexp)]
    [(pair? sexp)
     (case (car sexp)
       [(+) (add (parse (second sexp)) (parse (third sexp)))]
       [(-) (sub (parse (second sexp)) (parse (third sexp)))]
       [(fun) (fun (first (second sexp)) (parse (third sexp)))]
       [(if0) (if0 (parse (second sexp))
                   (parse (third sexp))
                   (parse (fourth sexp)))]
       [else (app (parse (first sexp)) (parse (second sexp)))])]))

(test (parse 3) (num 3))
(test (parse 'x) (id 'x))
(test (parse '{+ 1 2}) (add (num 1) (num 2)))
(test (parse '{- 1 2}) (sub (num 1) (num 2)))
(test (parse '{fun {x} x}) (fun 'x (id 'x)))
(test (parse '{1 2}) (app (num 1) (num 2)))
(test (parse '{if0 0 1 2}) (if0 (num 0) (num 1) (num 2)))

;; ----------------------------------------

;; compile : FAE CDefrdSub -> CFae
(define (compile a-fae ds)
  (type-case FAE a-fae
    [num (n) (cnum n)]
    [add (l r) (cadd (compile l ds) (compile r ds))]
    [sub (l r) (csub (compile l ds) (compile r ds))]
    [id (name) (cid (locate name ds))]
    [fun (param body-expr) (cfun 
                            (compile body-expr (aCSub param ds)))]
    [app (fun-expr arg-expr)
         (capp (compile fun-expr ds)
               (compile arg-expr ds))]
    [if0 (test-expr then-expr else-expr)
         (cif0 (compile test-expr ds)
               (compile then-expr ds)
               (compile else-expr ds))]))

(define (locate name ds)
  (type-case CDefrdSub ds
    [mtCSub () (error 'locate "free variable")]
    [aCSub (sub-name rest-ds)
           (if (symbol=? sub-name name)
               0
               (+ 1 (locate name rest-ds)))]))

;; ----------------------------------------

(define fae-reg (cnum 0))
(define ds-reg empty)

;; interp : -> void
(define (interp)
  (type-case CFAE fae-reg
    [cnum (n)
          (begin
            (set! v-reg (numV n))
            (continue))]
    [cadd (l r)
          (begin
            (set! fae-reg l)
            (set! k-reg (addSecondK r ds-reg k-reg))
            (interp))]
    [csub (l r)
          (begin
            (set! fae-reg l)
            (set! k-reg (subSecondK r ds-reg k-reg))
            (interp))]
    [cid (pos)
         (begin
           (set! v-reg (list-ref ds-reg pos))
           (continue))]
    [cfun (body-expr)
          (begin
            (set! v-reg (closureV body-expr ds-reg))
            (continue))]
    [capp (fun-expr arg-expr)
          (begin
            (set! fae-reg fun-expr)
            (set! k-reg (appArgK arg-expr ds-reg k-reg))
            (interp))]
    [cif0 (test-expr then-expr else-expr)
          (begin
            (set! fae-reg test-expr)
            (set! k-reg (doIfK then-expr else-expr ds-reg k-reg))
            (interp))]))

(define k-reg (mtK))
(define v-reg (numV 0))

;; continue : -> void
(define (continue)
  (type-case CFAE-Cont k-reg
    [mtK () v-reg]
    [addSecondK (r ds k)
                (begin
                  (set! fae-reg r)
                  (set! ds-reg ds)
                  (set! k-reg (doAddK v-reg k))
                  (interp))]
    [doAddK (v1 k)
            (begin
              (set! v-reg (num+ v1 v-reg))
              (set! k-reg k)
              (continue))]
    [subSecondK (r ds k)
                (begin
                  (set! fae-reg r)
                  (set! ds-reg ds)
                  (set! k-reg (doSubK v-reg k))
                  (interp))]
    [doSubK (v1 k)
            (begin
              (set! v-reg (num- v1 v-reg))
              (set! k-reg k)
              (continue))]
    [appArgK (arg-expr ds k)
             (begin
               (set! fae-reg arg-expr)
               (set! ds-reg ds)
               (set! k-reg (doAppK v-reg k))
               (interp))]
    [doAppK (fun-val k)
            (begin
              (set! fae-reg (closureV-body fun-val))
              (set! ds-reg (cons v-reg
                                 (closureV-ds fun-val)))
              (set! k-reg k)
              (interp))]
    [doIfK (then-expr else-expr ds k)
           (begin
             (if (numzero? v-reg)
                 (set! fae-reg then-expr)
                 (set! fae-reg else-expr))
             (set! ds-reg ds)
             (set! k-reg k)
             (interp))]))

;; num-op : (number number -> number) -> (FAE-Value FAE-Value -> FAE-Value)
(define (num-op op op-name)
  (lambda (x y)
    (numV (op (numV-n x) (numV-n y)))))

(define num+ (num-op + '+))
(define num- (num-op - '-))

;; numzero? : FAE-Value -> boolean
(define (numzero? x)
  (zero? (numV-n x)))

;; ----------------------------------------

(define init-k (mtK))
(define (interpx a-fae ds k)
  (begin
    (set! fae-reg a-fae)
    (set! ds-reg ds)
    (set! k-reg k)
    (interp)))

(test (interpx (compile (num 10) (mtCSub))
              empty
              init-k)
      (numV 10))
(test (interpx (compile (add (num 10) (num 7)) (mtCSub))
              empty
              init-k)
      (numV 17))
(test (interpx (compile (sub (num 10) (num 7)) (mtCSub))
              empty
              init-k)
      (numV 3))
(test (interpx (compile
               (app (fun 'x (add (id 'x) (num 12)))
                    (add (num 1) (num 17)))
               (mtCSub))
              empty
              init-k)
      (numV 30))
(test (interpx (compile (id 'x)
                       (aCSub 'x (mtCSub)))
              (list (numV 10))
              init-k)
      (numV 10))

(test (interpx (compile (app (fun 'x (add (id 'x) (num 12)))
                            (add (num 1) (num 17)))
                       (mtCSub))
              empty
              init-k)
      (numV 30))

(test (interpx (compile (app (fun 'x
                                 (app (fun 'f
                                           (add (app (id 'f) (num 1))
                                                (app (fun 'x
                                                          (app (id 'f) (num 2)))
                                                     (num 3))))
                                      (fun 'y (add (id 'x) (id 'y)))))
                            (num 0))
                       (mtCSub))
              empty
               init-k)
      (numV 3))

(test (interpx (compile (if0 (num 0)
                             (num 1)
                             (num 2))
                        (mtCSub))
               empty
               init-k)
      (numV 1))
(test (interpx (compile (if0 (num 1)
                             (num 0)
                             (num 2))
                        (mtCSub))
               empty
               init-k)
      (numV 2))
